from typing import Optional, Union

from selectolax.lexbor import LexborNode
from selectolax.parser import Node


def textof(
    node: Optional[Union[LexborNode, Node]],
    *,
    deep: bool = True,
    separator: str = "",
    strip: bool = False,
) -> str:
    return node.text(deep=deep, separator=separator, strip=strip) if node else ""


def some(node: Optional[Union[LexborNode, Node]]):
    return node if node else LexborNode()



import asyncio
from typing import List, Optional

from selectolax.lexbor import LexborHTMLParser

from .markdown import get_markdown
from .primp import Client, Response
from .schemas import (
    Aside,
    Flight,
    Lyrics,
    PartialWeatherForReport,
    Result,
    Snippet,
    Weather,
    WeatherForecast,
    Web,
)
from .utils import some, textof


user_agent = (
    "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/125.0.0.0 Safari/537.36 OPR/111.0.0.0"
)


def parse(res: Response) -> Result:
    parser = LexborHTMLParser(res.text)
    snippet = get_snippet(parser)
    aside = get_aside_block(parser)
    weather = get_weather(parser)
    web = get_web(parser)
    flights = get_flights(parser)
    lyrics = get_lyrics(parser)

    return Result(
        snippet=snippet,
        aside=aside,
        weather=weather,
        web=web,
        flights=flights,
        lyrics=lyrics,
    )


def get(q: str, hl: str, ua: Optional[str], **kwargs) -> Response:
    client = Client(impersonate="chrome_127", verify=False)
    res = client.get(
        "https://www.google.com/search",
        params={"q": q, "hl": hl, "client": "opera"},
        headers={"User-Agent": ua or user_agent},
        **kwargs,
    )
    assert res.status_code == 200, res.text
    return res


def search(q: str, *, hl: str = "en", ua: Optional[str] = None, **kwargs) -> Result:
    return parse(get(q, hl, ua, **kwargs))


async def asearch(
    q: str, *, hl: str = "en", ua: Optional[str] = None, **kwargs
) -> Result:
    res = await asyncio.to_thread(get, q, hl, ua, **kwargs)
    return await asyncio.to_thread(parse, res)


def get_snippet(parser: LexborHTMLParser) -> Optional[Snippet]:
    fsnippet_ele = parser.css_first(".xpdopen .hgKElc")

    # Get featured snippet (aka. quick answer)
    featured = (
        Snippet(
            text=get_markdown(fsnippet_ele.html or "", ".hgKElc"),
            highlighted=textof(fsnippet_ele.css_first("b"), deep=True, strip=True),
        )
        if fsnippet_ele
        else None
    )

    return featured


def get_aside_block(parser: LexborHTMLParser) -> Optional[Aside]:
    # Usually wikipedia blocks
    aside = parser.css_first(".xGj8Mb")
    if not aside:
        return None

    return Aside(text=aside.text(strip=True, separator=" ").replace("  ", " "))


def get_weather(parser: LexborHTMLParser) -> Optional[WeatherForecast]:
    # Get weather. Usually present when using the query "<place> weather"
    block = parser.css_first("#wob_wc")
    if not block:
        return None

    temp_c = textof(block.css_first("#wob_tm"))
    temp_f = textof(block.css_first("#wob_ttm"))

    precipitation = textof(block.css_first("#wob_pp"))
    humidity = textof(block.css_first("#wob_hm"))

    wind_metric = textof(block.css_first("#wob_ws"))
    wind_imperial = textof(block.css_first("#wob_tws"))

    description = textof(block.css_first("#wob_dc"))

    forecast = []
    for wth in block.css(".wob_df"):
        day = textof(wth.css_first(".Z1VzSb"))
        items = wth.css(".gNCp2e .wob_t")
        high_c = textof(items[0])
        high_f = textof(items[1])

        items1 = wth.css(".ZXCv8e .wob_t")
        low_c = textof(items1[0])
        low_f = textof(items1[1])

        forecast.append(
            PartialWeatherForReport(
                weekday=day,
                high_c=high_c,
                high_f=high_f,
                low_c=low_c,
                low_f=low_f,
            )
        )

    return WeatherForecast(
        now=Weather(
            c=temp_c,
            f=temp_f,
            precipitation=precipitation,
            humidity=humidity,
            wind_metric=wind_metric,
            wind_imperial=wind_imperial,
            description=description,
            forecast=forecast,
        ),
        warning=textof(block.css_first(".vk_h")) or None,
    )


def get_web(parser: LexborHTMLParser) -> List[Web]:
    # Get links
    items = []

    for item in parser.css(".N54PNb"):
        anchor = some(item.css_first("a")).attributes.get("href", "")
        title = textof(item.css_first("h3"), strip=True)

        items.append(
            Web(
                title=title,
                url=anchor,  # type: ignore
                text=textof(item.css_first(".VwiC3b")),
            )
        )

    return items


def get_flights(parser: LexborHTMLParser) -> List[Flight]:
    # Get flights when using the query "<place> to <place> [flights]"
    items = []

    for item in parser.css(".Ww4FFb.vt6azd .wyccme"):
        title = textof(item.css_first(".ZhosBf"), strip=True)
        description = textof(item.css_first(".GfzIoc"), strip=True)
        duration = textof(item.css_first(".TM2JYd"), strip=True)
        price = textof(item.css_first(".YK0p7d"), strip=True)

        items.append(
            Flight(
                title=title,
                description=description,
                duration=duration,
                price=price,
            )
        )

    return items


def get_lyrics(parser: LexborHTMLParser) -> Optional[Lyrics]:
    lyrics_ele = parser.css(".xaAUmb .ujudUb span")

    if not lyrics_ele:
        return None

    lyrics = []
    for ele in lyrics_ele:
        lyrics.append(textof(ele, strip=True))

    return Lyrics(
        text="\n".join(lyrics), is_partial=bool(parser.css_first(".xaAUmb .ujudUb a"))
    )



from typing import Dict, Optional, Tuple

class Client:
    """Initializes an HTTP client that can impersonate web browsers.

    Args:
        auth (tuple, optional): A tuple containing the username and password for basic authentication. Default is None.
        auth_bearer (str, optional): Bearer token for authentication. Default is None.
        params (dict, optional): Default query parameters to include in all requests. Default is None.
        headers (dict, optional): Default headers to send with requests. If `impersonate` is set, this will be ignored.
        cookies (dict, optional): - An optional map of cookies to send with requests as the `Cookie` header.
        timeout (float, optional): HTTP request timeout in seconds. Default is 30.
        cookie_store (bool, optional): Enable a persistent cookie store. Received cookies will be preserved and included
            in additional requests. Default is True.
        referer (bool, optional): Enable or disable automatic setting of the `Referer` header. Default is True.
        proxy (str, optional): Proxy URL for HTTP requests. Example: "socks5://127.0.0.1:9150". Default is None.
        impersonate (str, optional): Entity to impersonate. Example: "chrome_124". Default is None.
            Chrome: "chrome_99","chrome_100","chrome_101","chrome_104","chrome_105","chrome_106","chrome_107",
                "chrome_108","chrome_109","chrome_114","chrome_116","chrome_117","chrome_118","chrome_119",
                "chrome_120","chrome_123","chrome_124","chrome_126","chrome_127"
            Safari: "safari_ios_16.5","safari_ios_17.2","safari_ios_17.4.1","safari_15.3","safari_15.5","safari_15.6.1",
                "safari_16","safari_16.5","safari_17.2.1","safari_17.4.1","safari_17.5"
            OkHttp: "okhttp_3.9","okhttp_3.11","okhttp_3.13","okhttp_3.14","okhttp_4.9","okhttp_4.10","okhttp_5"
            Edge: "edge_99","edge_101","edge_122","edge_127"
        follow_redirects (bool, optional): Whether to follow redirects. Default is True.
        max_redirects (int, optional): Maximum redirects to follow. Default 20. Applies if `follow_redirects` is True.
        verify (bool, optional): Verify SSL certificates. Default is True.
        http1 (bool, optional): Use only HTTP/1.1. Default is None.
        http2 (bool, optional): Use only HTTP/2. Default is None.

    """
    def __init__(
        self,
        auth: Optional[Tuple[str, Optional[str]]] = None,
        auth_bearer: Optional[str] = None,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None,
        timeout: Optional[float] = 30,
        cookie_store: Optional[bool] = True,
        referer: Optional[bool] = True,
        proxy: Optional[str] = None,
        impersonate: Optional[str] = None,
        follow_redirects: Optional[bool] = True,
        max_redirects: Optional[int] = 20,
        verify: Optional[bool] = True,
        http1: Optional[bool] = None,
        http2: Optional[bool] = None,
    ) -> None: ...

    # === Methods ===
    def get(
        self,
        url: str,
        params: Optional[Dict[str, str]] = None,
        headers: Optional[Dict[str, str]] = None,
        cookies: Optional[Dict[str, str]] = None,
        auth: Optional[Tuple[str, Optional[str]]] = None,
        auth_bearer: Optional[str] = None,
        timeout: Optional[float] = 30,
    ) -> "Response":
        """Performs a GET request to the specified URL.

        Args:
            url (str): The URL to which the request will be made.
            params (Optional[Dict[str, str]]): A map of query parameters to append to the URL. Default is None.
            headers (Optional[Dict[str, str]]): A map of HTTP headers to send with the request. Default is None.
            cookies (Optional[Dict[str, str]]): - An optional map of cookies to send with requests as the `Cookie` header.
            auth (Optional[Tuple[str, Optional[str]]]): A tuple containing the username and an optional password
                for basic authentication. Default is None.
            auth_bearer (Optional[str]): A string representing the bearer token for bearer token authentication. Default is None.
            timeout (Optional[float]): The timeout for the request in seconds. Default is 30.
        """

class Response:
    content: str
    text: str
    status_code: int



from selectolax.parser import HTMLParser, Node

from .utils import textof


def get_markdown(html: str, target: str) -> str:
    """Gets Markdown.

    It may not look good if not formatted.

    Args:
        html (str): HTML string.
        target (str): CSS selector to the target. Used to unwrap the target.
    """
    parser = HTMLParser(html)
    n = parse_node_to_markdown(parser.css_first(target))
    n.unwrap()
    return parser.text().replace("\n\n\n", "\n\n").strip()


def parse_node_to_markdown(node: Node) -> Node:
    node.merge_text_nodes()
    replace_headings(node)

    for img in node.css("img"):
        img.replace_with("")

    for b in node.css("b"):
        b.replace_with(f"**{textof(b)}**")

    for br in node.css("br"):
        br.replace_with("\n\n")

    for p in node.css("p"):
        p.unwrap()

    for a in node.css("a"):
        a.replace_with(f"[{textof(a)}]({a.attributes.get('href', '/')})")

    for ul in node.css("ul"):
        texts = []
        for li in ul.css("li"):
            texts.append(f"- {textof(li, strip=True)}")

        ul.replace_with("\n".join(texts))

    for ol in node.css("ol"):
        texts = []
        for i, item in enumerate(ol.css("li")):
            texts.append(f"{i + 1}. {textof(item, strip=True)}")

        ol.replace_with("\n".join(texts))

    return node


def replace_headings(node: Node) -> None:
    for target in {"h1", "h2", "h3", "h4", "h5", "h6"}:
        for heading in node.css(target):
            heading.replace_with(f"{'#' * int(target[1])} {textof(heading)}")



from dataclasses import dataclass
from typing import List, Optional


@dataclass
class Snippet:
    text: str
    highlighted: Optional[str] = None


@dataclass
class Aside:
    text: str


@dataclass
class PartialWeatherForReport:
    weekday: str
    high_c: str
    high_f: str
    low_c: str
    low_f: str


@dataclass
class Weather:
    c: str
    f: str
    precipitation: str
    humidity: str
    wind_metric: str
    wind_imperial: str
    description: str
    forecast: List["PartialWeatherForReport"]


@dataclass
class WeatherForecast:
    now: Weather
    warning: Optional[str] = None


@dataclass
class Web:
    title: str
    url: str
    text: str


@dataclass
class Flight:
    title: str
    description: str
    duration: str
    price: str


@dataclass
class Lyrics:
    text: str
    is_partial: bool


@dataclass
class Result:
    snippet: Optional[Snippet]
    aside: Optional[Aside]
    weather: Optional[WeatherForecast]
    web: List[Web]
    flights: List[Flight]
    lyrics: Optional[Lyrics]
